#include "tree_structurer.hpp"
#include <algorithm>
#include <filesystem>
#include <fstream>
#include <sstream>
#include <stdexcept>

namespace fs = std::filesystem;

bool TreeStructurer::should_ignore_dir(const std::string& dirname) {
    static const std::vector<std::string> ignore_list = {
        "build", "venv", "myenv", "dist", "node_modules", "CMakeFiles",
        ".git", ".idea", ".vscode", "__pycache__", "**pycache**"
    };

    if (!dirname.empty() && (dirname == "__main__.py" || dirname == "__init__.py")) {
        return false;
    }

    if (std::find(ignore_list.begin(), ignore_list.end(), dirname) != ignore_list.end()) {
            return true;
        }

    if (!dirname.empty()) {
        if (dirname[0] == '.' || dirname[0] == '_') {
            return true;
        }
        if (dirname.find("__") == 0 && dirname.find("__", 2) != std::string::npos) {
        return true;
    }
    }

    return false;
}

bool TreeStructurer::should_ignore_file(const std::string& filename) {
    static const std::vector<std::string> ignore_extensions = {
        ".pyc", ".pyo", ".pyd", ".so", ".dll", ".dylib",
        ".o", ".obj", ".a", ".lib"
    };

    if (!filename.empty() && (filename == "__main__.py" || filename == "__init__.py")) {
        return false;
}

    if (!filename.empty() && (filename[0] == '.' || filename[0] == '_')) {
        return true;
    }

    fs::path path(filename);
    std::string ext = path.extension().string();
    if (std::find(ignore_extensions.begin(), ignore_extensions.end(), ext) != ignore_extensions.end()) {
        return true;
    }

    return false;
}

std::string TreeStructurer::get_relative_path(const fs::path& path, const fs::path& base) {
    fs::path rel = fs::relative(path, base);
    return rel.string();
}

std::vector<fs::path> TreeStructurer::get_filtered_paths(const fs::path& start) {
    std::vector<fs::path> paths;
    fs::directory_options options = fs::directory_options::skip_permission_denied;

    try {
        if (!fs::exists(start)) {
            throw std::runtime_error("Directory does not exist: " + start.string());
        }
        
        if (!fs::is_directory(start)) {
            throw std::runtime_error("Path is not a directory: " + start.string());
        }

        paths.push_back(start);

        bool is_empty = fs::directory_iterator(start) == fs::directory_iterator();
        if (is_empty) {
            throw std::runtime_error("Directory is empty: " + start.string());
        }

        for (const auto& entry : fs::recursive_directory_iterator(start, options)) {
            const auto& path = entry.path();

            bool should_skip = false;
            for (const auto& component : path) {
                if (should_ignore_dir(component.string())) {
                    should_skip = true;
                    break;
                }
            }
            if (should_skip) continue;

            if (entry.is_directory()) {
                if (!should_ignore_dir(path.filename().string())) {
                    paths.push_back(path);
                }
            } else {
                if (!should_ignore_file(path.filename().string())) {
                    paths.push_back(path);
                }
            }
        }
    } catch (const fs::filesystem_error& e) {
        throw std::runtime_error("Error accessing path: " + std::string(e.what()));
    }

    std::sort(paths.begin(), paths.end());
    return paths;
}

std::vector<std::string> TreeStructurer::get_directory_structure(const std::string& startpath) {
    std::vector<std::string> result;
    
    std::string normalized_path = startpath;
    if (startpath.substr(0, 2) == ".\\" || startpath.substr(0, 2) == "./") {
        normalized_path = startpath.substr(2);
    }
    
    fs::path start = normalized_path.empty() ? fs::current_path() : fs::path(normalized_path);

    try {
        auto paths = get_filtered_paths(start);
        if (paths.empty()) {
            throw std::runtime_error("No valid files or directories found in: " + start.string());
        }
        std::vector<bool> is_last_at_level(256, false);
        
        for (size_t i = 1; i < paths.size(); ++i) {
            const auto& path = paths[i];
            std::string rel_path = get_relative_path(path, start);
            
            int level = std::count(rel_path.begin(), rel_path.end(), fs::path::preferred_separator);

            bool is_last = true;
            for (size_t j = i + 1; j < paths.size(); ++j) {
                std::string next_rel_path = get_relative_path(paths[j], start);
                int next_level = std::count(next_rel_path.begin(), next_rel_path.end(), fs::path::preferred_separator);
                if (next_level == level) {
                    is_last = false;
                    break;
                }
                if (next_level < level) {
                    break;
                }
            }

            is_last_at_level[level] = is_last;

            std::string line;
            for (int j = 0; j < level; ++j) {
                if (j == level - 1) {
                    line += is_last ? "└── " : "├── ";
                } else {
                    line += is_last_at_level[j] ? "    " : "│   ";
                }
            }

            line += path.filename().string();
            if (fs::is_directory(path)) {
                line += "/";
            }
            result.push_back(line);
        }
    } catch (const fs::filesystem_error& e) {
        throw std::runtime_error("Failed to access directory: " + std::string(e.what()));
    }

    return result;
}

void TreeStructurer::create_structure_from_file(const std::string& filepath, const std::string& target_path) {
    std::vector<std::string> lines = read_structure_file(filepath);
    validate_structure(lines);
    TreeNode root = build_tree_from_lines(lines);
    create_node(root, target_path);
}

void TreeStructurer::create_structure_from_string(const std::string& structure, const std::string& target_path) {
    std::istringstream iss(structure);
    std::vector<std::string> lines;
    std::string line;
    
    while (std::getline(iss, line)) {
        if (!line.empty()) {
            lines.push_back(line);
        }
    }
    
    validate_structure(lines);
    TreeNode root = build_tree_from_lines(lines);
    create_node(root, target_path);
}

void TreeStructurer::validate_structure(const std::vector<std::string>& lines) {
    if (lines.empty()) {
        throw std::runtime_error("Empty structure provided");
    }
    
    int prev_indent = -1;
    for (const auto& line : lines) {
        int current_indent = get_indent_level(line);
        if (current_indent > prev_indent + 1) {
            throw std::runtime_error("Invalid indentation level in structure");
        }
        prev_indent = current_indent;
    }
}


size_t TreeStructurer::get_indent_level(const std::string& line) {
    size_t level = 0;
    for (size_t i = 0; i < line.length(); ++i) {
        if (line[i] == ' ') {
            continue;
        }
        if (line[i] == '-' || 
            static_cast<unsigned char>(line[i]) == INDENT_MARKER_PIPE ||
            static_cast<unsigned char>(line[i]) == INDENT_MARKER_DASH ||
            static_cast<unsigned char>(line[i]) == INDENT_MARKER_END) {
            level++;
        } else {
            break;
        }
    }
    return level / 4;
}


TreeStructurer::TreeNode TreeStructurer::build_tree_from_lines(const std::vector<std::string>& lines) {
    TreeNode root{"root", false, {}};
    std::vector<TreeNode*> stack{&root};
    size_t prev_indent = 0;

    for (const auto& line : lines) {
        if (line.empty()) continue;

        size_t current_indent = get_indent_level(line);

        while (current_indent <= prev_indent && stack.size() > 1) {
            stack.pop_back();
            prev_indent--;
        }

        TreeNode new_node = parse_structure_line(line, current_indent);
        if (!new_node.name.empty()) {
            stack.back()->children.push_back(new_node);

            if (!new_node.is_file) {
                stack.push_back(&(stack.back()->children.back()));
            }

            prev_indent = current_indent;
        }
    }

    return root;
}


void TreeStructurer::create_node(const TreeNode& node, const std::filesystem::path& current_path) {
    std::filesystem::path new_path = current_path / node.name;
    
    if (node.is_file) {
        create_file(new_path);
    } else {
        create_directory(new_path);
        for (const auto& child : node.children) {
            create_node(child, new_path);
        }
    }
}

void TreeStructurer::create_directory(const std::filesystem::path& path) {
    if (!std::filesystem::exists(path)) {
        if (!std::filesystem::create_directory(path)) {
            throw std::runtime_error("Failed to create directory: " + path.string());
        }
    }
}

void TreeStructurer::create_file(const std::filesystem::path& path) {
    if (!std::filesystem::exists(path)) {
        std::ofstream file(path);
        if (!file.is_open()) {
            throw std::runtime_error("Failed to create file: " + path.string());
        }
    }
}

TreeStructurer::TreeNode TreeStructurer::parse_structure_line(const std::string& line, size_t indent_level) {
    size_t name_start = 0;
    for (size_t i = 0; i < line.length(); ++i) {
        if (line[i] != ' ' && 
            static_cast<unsigned char>(line[i]) != INDENT_MARKER_PIPE &&
            static_cast<unsigned char>(line[i]) != INDENT_MARKER_DASH &&
            static_cast<unsigned char>(line[i]) != INDENT_MARKER_END &&
            line[i] != '-' &&
            line[i] != '├' &&
            line[i] != '└') {
            name_start = i;
            break;
        }
    }
    
    std::string name = line.substr(name_start);
    name.erase(0, name.find_first_not_of(" "));
    name.erase(name.find_last_not_of(" ") + 1);
    
    bool is_file = name.empty() || name.back() != '/';
    
    if (!is_file) {
        name.pop_back();
    }
    
    return TreeNode{name, is_file, {}};
}

bool TreeStructurer::is_directory_marker(const std::string& line) {
    return !line.empty() && line.back() == DIRECTORY_MARKER;
}

std::string TreeStructurer::create_indent(size_t level) {
    return std::string(level * 4, ' ');
}

std::vector<std::string> TreeStructurer::read_structure_file(const std::string& filepath) {
    std::vector<std::string> lines;
    std::ifstream file(filepath);
    
    if (!file.is_open()) {
        throw std::runtime_error("Failed to open file: " + filepath);
    }
    
    std::string line;
    while (std::getline(file, line)) {
        if (!line.empty()) {
            lines.push_back(line);
        }
    }
    
    return lines;
}